/******************************************************************
* dft.c - An attempted DFT implementation
*  
* 		Convert time -> frequency and freq-> time :D
*
* Created by:	isabella, 2011.09.26 - 11:57:42am
* 
******************************************************************/


#include "dft.h"
#include "memory.h"

// Initialize partial to given values
dftRet_t	dftInitPartial(partialPtr_t pPartial, int sampleRate, size_t sampleWindowSize, float freq) {
	// Set freq
	pPartial->freq 		= freq;

	// Most recent calculated values 
	pPartial->phase 	= 0.0f;
	pPartial->amp 		= 0.0f;

	// Stores running 
	pPartial->aN 		= 0.0f;
	pPartial->bN 		= 0.0f;
	// Number of samples represented by raw a/b pairs
	pPartial->aBCount	= 0;

	// print current freq
	printf("%3.02fHz  Pre-rendering sin/cos table...", pPartial->freq);

	// Calculate how many samples it takes for one period
	pPartial->period 	= (float)sampleRate / freq;
	// 
	pPartial->sinfTblSize 	= (size_t)(((float)sampleWindowSize / pPartial->period) * (pPartial->period + 2));

	// Allocate pre-rendered table
	pPartial->pSinfTbl 	= (float *)calloc(pPartial->sinfTblSize, sizeof(float));

	// for now do cosine too (tho sine can be used for this...)
	pPartial->pCosfTbl 	= (float *)calloc(pPartial->sinfTblSize, sizeof(float));
	
	if(!pPartial->pSinfTbl || !pPartial->pCosfTbl) {
		printf(" Failed!~\n");
		return(DFT_ERROR);
	}

	// Pre-render sinf table
	const float	freqScaler = dftFreq(freq, sampleRate);
	unsigned int	ndx;
	for(ndx=0; ndx < pPartial->sinfTblSize; ndx++) {
		pPartial->pSinfTbl[ndx] = sin(freqScaler * (float)ndx);
		pPartial->pCosfTbl[ndx] = cos(freqScaler * (float)ndx);
	};
	
	// zero first sample of B part (B part not added until after first sample)
//	pPartial->pSinfTbl[0] = 0.0f;

	// Done with pre-rendering :D
	printf(" Done!~\n");
	return(DFT_OK);
}


dftPtr_t	dftCreate(int sampleRate, size_t sampleWindowSize, float freqA, float freqB, float freqStep, dftCreateType_t dftType) {
	int count = ((freqB>freqA)?freqB-freqA:freqA-freqB)/freqStep;

	int mallocSize  = count*sizeof(partial_t)+sizeof(dft_t);

	dftPtr_t retVal = (dftPtr_t)malloc(mallocSize);

	if(retVal) {
		printf("Allocated space for %d entries.\n", count);

		memset(retVal, 0, mallocSize);

		count = 0;
		while(freqA < freqB) {
			// Initalize partial
			if(dftInitPartial(&retVal->list[count], sampleRate, sampleWindowSize, freqA) == DFT_ERROR) {
				dftDestroy(retVal);
				dftErr("Malloc Failed");
				return(NULL);
			}
			// Increment frequency
			switch(dftType) {
				case DFT_LINEAR:
					freqA += freqStep;
					break;
				case DFT_LOG:
				default:
					freqA *= freqStep;
					break;
			};
			count++;
		};

		retVal->count		 = count;
		retVal->sampleRate 	 = sampleRate;
		retVal->sampleWindowSize = sampleWindowSize;
//		retVal->dftType		 = DFT_WINDOWED;
		retVal->dftType		 = DFT_CONTINUOUS;
//		retVal->dftType		 = DFT_FAST;

		printf("Created %d partials, %0.03f ~ %0.03f [step: %0.03f]\n", count, freqA, freqB, freqStep);
	} else {
		dftErr("Malloc Failed");
	}
	return(retVal);
}

//
void	dftDestroy(dftPtr_t pDft) {
	unsigned int ndx;
	for(ndx = 0; ndx < pDft->count; ndx++) {
		// Free pre-rendered sinf
		free(pDft->list[ndx].pSinfTbl);	
		free(pDft->list[ndx].pCosfTbl);
	};
	// free pDft
	free(pDft);
}

// Convert from time domain (PCM Audio) to frequency domain
dftRet_t	dft2Freq(dftPtr_t pDft, dftSamplePtr_t sampleBuf) {
	unsigned int	partialNdx;
	for(partialNdx = 0; partialNdx < pDft->count; partialNdx++) {

		float	aN      = 0;
		float	bN      = 0;
		float	aBCount = 1.0f;

		partialPtr_t pPartial = &pDft->list[partialNdx];

		unsigned int sNdx;

		switch(pDft->dftType) {
			case DFT_FAST:
				// Fast!~ (Only compare first 2 cycles of at most. :p 
				// Phase calculation is prolly useless after here..
				aBCount = (size_t)(pPartial->period * 2.0f);
				if(aBCount > pDft->sampleWindowSize) {
					// if we don't have at 2 cycles worth of window, use what we can 'see'
					aBCount = pDft->sampleWindowSize;
				}

				for(sNdx = 0; sNdx < (unsigned int)aBCount; sNdx++) {
					float funcX = (float)sampleBuf[sNdx];

					aN += funcX * pPartial->pCosfTbl[sNdx];
					bN += funcX * pPartial->pSinfTbl[sNdx];
				}
				// Guessing at how to weight this >_> ;p
				aN = aN * (aBCount / pDft->sampleWindowSize);
				bN = bN * (aBCount / pDft->sampleWindowSize);
				break;

			case DFT_CONTINUOUS: {
				// Continuous (phase sync'd to wave period, consistant between windows)
				unsigned int samples;
				unsigned int tblNdx;

				if((size_t)pPartial->period > pDft->sampleWindowSize) {
					samples = pDft->sampleWindowSize;
					aBCount = pDft->sampleWindowSize;
					//break;
				} else {
					// Rounded off count of samples we're using (aBCount)
					aBCount = (float)((unsigned int)(((float)(pDft->sampleWindowSize + pPartial->aBCount) / pPartial->period)) * pPartial->period);
					// Number of samples in this window to calculate
					samples = (unsigned int)aBCount - pPartial->aBCount;
				}

//if(partialNdx < 2)
//printf("freq:%0.02f p->aBcount %d, Samples %d, remainder %d\n", pPartial->freq, pPartial->aBCount, samples, pDft->sampleWindowSize - samples);

				// Start with previous values (if any)
				aN = pPartial->aN;
				bN = pPartial->bN;

				tblNdx = pPartial->aBCount;

				for(sNdx = 0; sNdx < samples; sNdx++) {
					float funcX = (float)sampleBuf[sNdx];

					aN += funcX * pPartial->pCosfTbl[tblNdx];
					bN += funcX * pPartial->pSinfTbl[tblNdx];
					tblNdx++;
				}

				float aN_2 = 0.0f;
				float bN_2 = 0.0f;

				tblNdx = 0; 

				for(sNdx = samples; sNdx < pDft->sampleWindowSize; sNdx++) {
					float funcX = (float)sampleBuf[sNdx];

					aN_2 += funcX * pPartial->pCosfTbl[tblNdx];
					bN_2 += funcX * pPartial->pSinfTbl[tblNdx];
					tblNdx++;
				}

				pPartial->aN = aN_2;
				pPartial->bN = bN_2;

				pPartial->aBCount = pDft->sampleWindowSize - samples;
				break;
			}
			case DFT_WINDOWED:
			default:
				// Windowed (Calculate over the given window only. Phase sync'd to start of window)
				for(sNdx = 0; sNdx < pDft->sampleWindowSize; sNdx++) {
					float funcX = (float)sampleBuf[sNdx];

					aN += funcX * pPartial->pCosfTbl[sNdx];
					bN += funcX * pPartial->pSinfTbl[sNdx];
				}
				aBCount = (float)pDft->sampleWindowSize;
				break;
		}

		if(aBCount >= 1.0f) {
			pPartial->amp 	= 2.0f * sqrtf(aN * aN + bN * bN) / aBCount;
			pPartial->phase	= (bN != 0.0f) ? atanf(aN/bN) : 0.0f;
		}/* else {
			pPartial->amp 	= 0.0f;
			pPartial->phase	= 0.0f;
		}*/

		if(pPartial->amp > pPartial->peakAmp) {
			pPartial->peakAmp = pPartial->amp;
		}

//		printf("[%3d] freq: %0.0003f Amp: %0.00004f Phase: %0.00004f\n", partialNdx, pDft->list[partialNdx].freq, pDft->list[partialNdx].amp, pDft->list[partialNdx].phase);	
	}
	return(DFT_OK);
}

// Convert from frequency to time domain
dftRet_t	dft2PCM(dftPtr_t pDft, dftSamplePtr_t sampleBuf) {
	unsigned partialNdx;
	memset(sampleBuf, 0x0, sizeof(dftSample_t) * pDft->sampleWindowSize);
	for(partialNdx = 0; partialNdx < pDft->count; partialNdx++) {
		float freqScaler = dftFreq(pDft->list[partialNdx].freq, pDft->sampleRate);
		unsigned int sNdx;
		for(sNdx=0; sNdx < pDft->sampleWindowSize; sNdx++) {
			sampleBuf[sNdx] += (dftSample_t)(sinf(freqScaler * (float)sNdx + pDft->list[partialNdx].phase) * pDft->list[partialNdx].amp);
			
		}		
	}	
	return(DFT_OK);
}

// Determine the freq/phase of a single frequency
dftRet_t	dft2FreqBand(int sampleRate, dftSamplePtr_t sampleBuf, size_t sampleWindowSize, float freq, float *amp, float *phase) {
	float	aN = 0;
	float	bN = 0;

	unsigned int sNdx;
	for(sNdx=0; sNdx < sampleWindowSize; sNdx++) {
			float funcX = (float)sampleBuf[sNdx];
			float freqBase = dftFreq(freq, sampleRate) * (float)sNdx;

			aN += funcX * cosf(freqBase);
			if(sNdx >= 1) {
				bN += funcX * sinf(freqBase);
			}
		
	}
	*amp = 2.0f * sqrtf(aN * aN + bN * bN) / (float)sampleWindowSize;
	*phase = atanf(aN/bN);	

	printf("[B] freq: %0.0003f Amp: %0.00004f Phase: %0.00004f\n", freq, *amp, *phase);	
	return(DFT_OK);
}


